import numpy as np
import rioxarray  # noqa
import xarray as xr
from affine import Affine

from rasterix import RasterIndex, assign_index


def test_rectilinear():
    source = "/vsicurl/https://noaadata.apps.nsidc.org/NOAA/G02135/south/daily/geotiff/2024/01_Jan/S_20240101_concentration_v3.0.tif"
    da_no_raster_index = xr.open_dataarray(source, engine="rasterio")
    da_raster_index = assign_index(da_no_raster_index)
    assert da_raster_index.equals(da_no_raster_index)


# TODO: parameterize over
# 1. y points up;
# 2. y points down
def test_sel_slice():
    ds = xr.Dataset({"foo": (("y", "x"), np.ones((10, 12)))})
    transform = Affine.identity()
    ds = ds.rio.write_transform(transform)
    ds = assign_index(ds)

    assert ds.xindexes["x"].transform() == transform

    actual = ds.sel(x=slice(4), y=slice(3, 5))
    assert isinstance(actual.xindexes["x"], RasterIndex)
    assert isinstance(actual.xindexes["y"], RasterIndex)
    actual_transform = actual.xindexes["x"].transform()

    assert actual_transform == actual.rio.transform()
    assert actual_transform == (transform * Affine.translation(0, 3))



# exactexact wrappers
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

import geopandas as gpd
import numpy as np
import xarray as xr
from exactextract import exact_extract
from exactextract.raster import NumPyRasterSource

from .utils import geometries_as_dask_array, is_in_memory

if TYPE_CHECKING:
    import dask.array
    import dask_geopandas

MIN_CHUNK_SIZE = 2  # exactextract cannot handle arrays of size 1.

__all__ = [
    "coverage",
]


def get_dtype(coverage_weight, geometries):
    if coverage_weight.lower() == "fraction":
        dtype = "float64"
    elif coverage_weight.lower() == "none":
        dtype = np.min_scalar_type(len(geometries))
    else:
        raise NotImplementedError
    return dtype


def np_coverage(
    x: np.ndarray,
    y: np.ndarray,
    *,
    geometries: gpd.GeoDataFrame,
    coverage_weight: Literal["fraction", "none"] = "fraction",
) -> np.ndarray[Any, Any]:
    """
    Parameters
    ----------

    """
    assert x.ndim == 1
    assert y.ndim == 1

    dtype = get_dtype(coverage_weight, geometries)

    xsize = x.size
    ysize = y.size

    # we need the top left corner, and the bottom right corner
    dx0 = (x[1] - x[0]) / 2
    dx1 = (x[-1] - x[-2]) / 2
    dy0 = np.abs(y[1] - y[0]) / 2
    dy1 = np.abs(y[-1] - y[-2]) / 2
    if y[0] > y[-1]:
        dy0, dy1 = dy1, dy0

    shape = (ysize, xsize)
    raster = NumPyRasterSource(
        np.broadcast_to([1], shape),
        xmin=x.min() - dx0,
        xmax=x.max() + dx1,
        ymin=y.min() - dy0,
        ymax=y.max() + dy1,
        srs_wkt=geometries.crs.to_wkt(),
    )
    result = exact_extract(
        rast=raster,
        vec=geometries,
        ops=["cell_id", f"coverage(coverage_weight={coverage_weight})"],
        output="pandas",
        # max_cells_in_memory=2*x.size * y.size
    )
    out = np.zeros((len(geometries), *shape), dtype=dtype)
    # TODO: vectorized assignment?
    for i in range(len(geometries)):
        res = result.loc[i]
        # indices = np.unravel_index(res.cell_id, shape=shape)
        # out[(i, *indices)] = offset + i + 1 # 0 is the fill value
        out[i, ...].flat[res.cell_id] = res.coverage
    return out


def coverage_np_dask_wrapper(
    x: np.ndarray, y: np.ndarray, geom_array: np.ndarray, coverage_weight, crs
) -> np.ndarray:
    return np_coverage(
        x=x,
        y=y,
        geometries=gpd.GeoDataFrame(geometry=geom_array, crs=crs),
        coverage_weight=coverage_weight,
    )


def dask_coverage(
    x: dask.array.Array,
    y: dask.array.Array,
    *,
    geom_array: dask.array.Array,
    coverage_weight: Literal["fraction", "none"] = "fraction",
    crs: Any,
) -> dask.array.Array:
    import dask.array

    if any(c == 1 for c in x.chunks) or any(c == 1 for c in y.chunks):
        raise ValueError("exactextract does not support a chunksize of 1. Please rechunk to avoid this")

    return dask.array.blockwise(
        coverage_np_dask_wrapper,
        "gji",
        x,
        "i",
        y,
        "j",
        geom_array,
        "g",
        crs=crs,
        coverage_weight=coverage_weight,
        dtype=get_dtype(coverage_weight, geom_array),
    )


def coverage(
    obj: xr.Dataset | xr.DataArray,
    geometries: gpd.GeoDataFrame | dask_geopandas.GeoDataFrame,
    *,
    xdim="x",
    ydim="y",
    coverage_weight="fraction",
) -> xr.DataArray:
    """
    Returns "coverage" fractions for each pixel for each geometry calculated using exactextract.

    Parameters
    ----------
    obj : xr.DataArray | xr.Dataset
        Xarray object used to extract the grid
    geometries: GeoDataFrame | DaskGeoDataFrame
        Geometries used for to calculate coverage
    xdim: str
        Name of the "x" dimension on ``obj``.
    ydim: str
        Name of the "y" dimension on ``obj``.
    coverage_weight: {"fraction", "none", "area_cartesian", "area_spherical_m2", "area_spherical_km2"}
        Weights to estimate, passed directly to exactextract.

    Returns
    -------
    DataArray
        3D dataarray with coverage fraction. The additional dimension is "geometry".
    """
    if "spatial_ref" not in obj.coords:
        raise ValueError("Xarray object must contain the `spatial_ref` variable.")
    # FIXME: assert obj.crs == geometries.crs
    if is_in_memory(obj=obj, geometries=geometries):
        out = np_coverage(
            x=obj[xdim].data,
            y=obj[ydim].data,
            geometries=geometries,
            coverage_weight=coverage_weight,
        )
        geom_array = geometries.to_numpy().squeeze(axis=1)
    else:
        from dask.array import from_array

        geom_dask_array = geometries_as_dask_array(geometries)
        out = dask_coverage(
            x=from_array(obj[xdim].data, chunks=obj.chunksizes.get(xdim, -1)),
            y=from_array(obj[ydim].data, chunks=obj.chunksizes.get(ydim, -1)),
            geom_array=geom_dask_array,
            crs=geometries.crs,
            coverage_weight=coverage_weight,
        )
        if isinstance(geometries, gpd.GeoDataFrame):
            geom_array = geometries.to_numpy().squeeze(axis=1)
        else:
            geom_array = geom_dask_array

    coverage = xr.DataArray(
        dims=("geometry", ydim, xdim),
        data=out,
        coords=xr.Coordinates(
            coords={
                xdim: obj.coords[xdim],
                ydim: obj.coords[ydim],
                "spatial_ref": obj.spatial_ref,
                "geometry": geom_array,
            },
            indexes={xdim: obj.xindexes[xdim], ydim: obj.xindexes[ydim]},
        ),
    )
    return coverage



import dask_geopandas as dgpd
import geodatasets
import geopandas as gpd
import rioxarray  # noqa
import xarray as xr
from xarray.tests import raise_if_dask_computes

from rasterix.rasterize.exact import coverage
from rasterix.rasterize.rasterio import rasterize


def test_coverage():
    ds = xr.tutorial.open_dataset("eraint_uvz")
    ds = ds.rio.write_crs("epsg:4326")
    world = gpd.read_file(geodatasets.get_path("naturalearth land"))

    rasterized = coverage(ds, world[["geometry"]], xdim="longitude", ydim="latitude")

    chunked = ds.chunk(latitude=119, longitude=-1)
    with raise_if_dask_computes():
        drasterized = coverage(chunked, world[["geometry"]], xdim="longitude", ydim="latitude")
    xr.testing.assert_allclose(rasterized, drasterized)

    dask_geoms = dgpd.from_geopandas(world, chunksize=5)
    with raise_if_dask_computes():
        drasterized = coverage(chunked, dask_geoms[["geometry"]], xdim="longitude", ydim="latitude")
    xr.testing.assert_allclose(rasterized, drasterized)


def test_rasterize():
    ds = xr.tutorial.open_dataset("eraint_uvz")
    ds = ds.rio.write_crs("epsg:4326")
    world = gpd.read_file(geodatasets.get_path("naturalearth land"))

    rasterized = rasterize(ds, world[["geometry"]], xdim="longitude", ydim="latitude")

    chunked = ds.chunk(latitude=119, longitude=-1)
    with raise_if_dask_computes():
        drasterized = rasterize(chunked, world[["geometry"]], xdim="longitude", ydim="latitude")
    xr.testing.assert_identical(rasterized, drasterized)

    dask_geoms = dgpd.from_geopandas(world, chunksize=5)
    with raise_if_dask_computes():
        drasterized = rasterize(chunked, dask_geoms[["geometry"]], xdim="longitude", ydim="latitude")
    xr.testing.assert_identical(rasterized, drasterized)



from __future__ import annotations

import textwrap
from collections.abc import Hashable, Mapping
from typing import Any, TypeVar

import numpy as np
import pandas as pd
from affine import Affine
from xarray import Coordinates, DataArray, Dataset, Index, Variable
from xarray.core.coordinate_transform import CoordinateTransform

# TODO: import from public API once it is available
from xarray.core.indexes import CoordinateTransformIndex, PandasIndex
from xarray.core.indexing import IndexSelResult, merge_sel_results

from rasterix.rioxarray_compat import guess_dims

T_Xarray = TypeVar("T_Xarray", "DataArray", "Dataset")


def assign_index(obj: T_Xarray, *, x_dim: str | None = None, y_dim: str | None = None) -> T_Xarray:
    if x_dim is None or y_dim is None:
        guessed_x, guessed_y = guess_dims(obj)
    x_dim = x_dim or guessed_x
    y_dim = y_dim or guessed_y

    index = RasterIndex.from_transform(
        obj.rio.transform(), obj.sizes[x_dim], obj.sizes[y_dim], x_dim=x_dim, y_dim=y_dim
    )
    coords = Coordinates.from_xindex(index)
    return obj.assign_coords(coords)


class AffineTransform(CoordinateTransform):
    """Affine 2D transform wrapper."""

    affine: Affine
    xy_dims: tuple[str, str]

    def __init__(
        self,
        affine: Affine,
        width: int,
        height: int,
        x_coord_name: Hashable = "xc",
        y_coord_name: Hashable = "yc",
        x_dim: str = "x",
        y_dim: str = "y",
        dtype: Any = np.dtype(np.float64),
    ):
        super().__init__((x_coord_name, y_coord_name), {x_dim: width, y_dim: height}, dtype=dtype)
        self.affine = affine

        # array dimensions in reverse order (y = rows, x = cols)
        self.xy_dims = self.dims[0], self.dims[1]
        self.dims = self.dims[1], self.dims[0]

    def forward(self, dim_positions):
        positions = tuple(dim_positions[dim] for dim in self.xy_dims)
        x_labels, y_labels = self.affine * positions

        results = {}
        for name, labels in zip(self.coord_names, [x_labels, y_labels]):
            results[name] = labels

        return results

    def reverse(self, coord_labels):
        labels = tuple(coord_labels[name] for name in self.coord_names)
        x_positions, y_positions = ~self.affine * labels

        results = {}
        for dim, positions in zip(self.xy_dims, [x_positions, y_positions]):
            results[dim] = positions

        return results

    def equals(self, other):
        if not isinstance(other, AffineTransform):
            return False
        return self.affine == other.affine and self.dim_size == other.dim_size


class AxisAffineTransform(CoordinateTransform):
    """Axis-independent wrapper of an affine 2D transform with no skew/rotation."""

    affine: Affine
    is_xaxis: bool
    coord_name: Hashable
    dim: str
    size: int

    def __init__(
        self,
        affine: Affine,
        size: int,
        coord_name: Hashable,
        dim: str,
        is_xaxis: bool,
        dtype: Any = np.dtype(np.float64),
    ):
        assert affine.is_rectilinear and (affine.b == affine.d == 0)

        super().__init__((coord_name,), {dim: size}, dtype=dtype)
        self.affine = affine
        self.is_xaxis = is_xaxis
        self.coord_name = coord_name
        self.dim = dim
        self.size = size

    def forward(self, dim_positions: dict[str, Any]) -> dict[Hashable, Any]:
        positions = np.asarray(dim_positions[self.dim])

        if self.is_xaxis:
            labels, _ = self.affine * (positions, np.zeros_like(positions))
        else:
            _, labels = self.affine * (np.zeros_like(positions), positions)

        return {self.coord_name: labels}

    def reverse(self, coord_labels: dict[Hashable, Any]) -> dict[str, Any]:
        labels = np.asarray(coord_labels[self.coord_name])

        if self.is_xaxis:
            positions, _ = ~self.affine * (labels, np.zeros_like(labels))
        else:
            _, positions = ~self.affine * (np.zeros_like(labels), labels)

        return {self.dim: positions}

    def equals(self, other):
        if not isinstance(other, AxisAffineTransform):
            return False

        # only compare the affine parameters of the relevant axis
        if self.is_xaxis:
            affine_match = self.affine.a == other.affine.a and self.affine.c == other.affine.c
        else:
            affine_match = self.affine.e == other.affine.e and self.affine.f == other.affine.f

        return affine_match and self.size == other.size

    def generate_coords(self, dims: tuple[str, ...] | None = None) -> dict[Hashable, Any]:
        assert dims is None or dims == self.dims
        return self.forward({self.dim: np.arange(self.size)})

    def slice(self, slice: slice) -> AxisAffineTransform:
        start = max(slice.start or 0, 0)
        stop = min(slice.stop or self.size, self.size)
        step = slice.step or 1

        # TODO: support reverse transform (i.e., start > stop)?
        assert start < stop

        size = (stop - start) // step
        scale = float(step)

        if self.is_xaxis:
            affine = self.affine * Affine.translation(start, 0.0) * Affine.scale(scale, 1.0)
        else:
            affine = self.affine * Affine.translation(0.0, start) * Affine.scale(1.0, scale)

        return type(self)(
            affine,
            size,
            self.coord_name,
            self.dim,
            is_xaxis=self.is_xaxis,
            dtype=self.dtype,
        )


class AxisAffineTransformIndex(CoordinateTransformIndex):
    """Axis-independent Xarray Index for an affine 2D transform with no
    skew/rotation.

    For internal use only.

    This Index class provides specific behavior on top of
    Xarray's `CoordinateTransformIndex`:

    - Data slicing computes a new affine transform and returns a new
      `AxisAffineTransformIndex` object

    - Otherwise data selection creates and returns a new Xarray
      `PandasIndex` object for non-scalar indexers

    - The index can be converted to a `pandas.Index` object (useful for Xarray
      operations that don't work with Xarray indexes yet).

    """

    axis_transform: AxisAffineTransform
    dim: str

    def __init__(self, transform: AxisAffineTransform):
        assert isinstance(transform, AxisAffineTransform)
        super().__init__(transform)
        self.axis_transform = transform
        self.dim = transform.dim

    def isel(  # type: ignore[override]
        self, indexers: Mapping[Any, int | slice | np.ndarray | Variable]
    ) -> AxisAffineTransformIndex | PandasIndex | None:
        idxer = indexers[self.dim]

        # generate a new index with updated transform if a slice is given
        if isinstance(idxer, slice):
            return AxisAffineTransformIndex(self.axis_transform.slice(idxer))
        # no index for vectorized (fancy) indexing with n-dimensional Variable
        elif isinstance(idxer, Variable) and idxer.ndim > 1:
            return None
        # no index for scalar value
        elif np.ndim(idxer) == 0:
            return None
        # otherwise return a PandasIndex with values computed by forward transformation
        else:
            values = self.axis_transform.forward({self.dim: idxer})[self.axis_transform.coord_name]
            if isinstance(idxer, Variable):
                new_dim = idxer.dims[0]
            else:
                new_dim = self.dim
            return PandasIndex(values, new_dim, coord_dtype=values.dtype)

    def sel(self, labels, method=None, tolerance=None):
        coord_name = self.axis_transform.coord_name
        label = labels[coord_name]

        if isinstance(label, slice):
            if label.start is None:
                label = slice(0, label.stop, label.step)
            if label.step is None:
                # continuous interval slice indexing (preserves the index)
                pos = self.transform.reverse({coord_name: np.array([label.start, label.stop])})
                # np.round rounds to even, this way we round upwards
                pos = np.floor(pos[self.dim] + 0.5).astype("int")
                new_start = max(pos[0], 0)
                new_stop = min(pos[1], self.axis_transform.size)
                return IndexSelResult({self.dim: slice(new_start, new_stop)})
            else:
                # otherwise convert to basic (array) indexing
                label = np.arange(label.start, label.stop, label.step)

        # support basic indexing (in the 1D case basic vs. vectorized indexing
        # are pretty much similar)
        unwrap_xr = False
        if not isinstance(label, Variable | DataArray):
            # basic indexing -> either scalar or 1-d array
            try:
                var = Variable("_", label)
            except ValueError:
                var = Variable((), label)
            labels = {self.dim: var}
            unwrap_xr = True

        result = super().sel(labels, method=method, tolerance=tolerance)

        if unwrap_xr:
            dim_indexers = {self.dim: result.dim_indexers[self.dim].values}
            result = IndexSelResult(dim_indexers)

        return result

    def to_pandas_index(self) -> pd.Index:
        import pandas as pd

        values = self.transform.generate_coords()
        return pd.Index(values[self.dim])


# The types of Xarray indexes that may be wrapped by RasterIndex
WrappedIndex = AxisAffineTransformIndex | PandasIndex | CoordinateTransformIndex
WrappedIndexCoords = Hashable | tuple[Hashable, Hashable]


def _filter_dim_indexers(index: WrappedIndex, indexers: Mapping) -> Mapping:
    if isinstance(index, CoordinateTransformIndex):
        dims = index.transform.dims
    else:
        # PandasIndex
        dims = (str(index.dim),)

    return {dim: indexers[dim] for dim in dims if dim in indexers}


class RasterIndex(Index):
    """Xarray index for raster coordinates.

    RasterIndex is itself a wrapper around one or more Xarray indexes associated
    with either the raster x or y axis coordinate or both, depending on the
    affine transformation and prior data selection (if any):

    - The affine transformation is not rectilinear or has rotation: this index
      encapsulates a single `CoordinateTransformIndex` object for both the x and
      y axis (2-dimensional) coordinates.

    - The affine transformation is rectilinear ands has no rotation: this index
      encapsulates one or two index objects for either the x or y axis or both
      (1-dimensional) coordinates. The index type is either a subclass of
      `CoordinateTransformIndex` that supports slicing or `PandasIndex` (e.g.,
      after data selection at arbitrary locations).

    """

    _wrapped_indexes: dict[WrappedIndexCoords, WrappedIndex]

    def __init__(self, indexes: Mapping[WrappedIndexCoords, WrappedIndex]):
        idx_keys = list(indexes)
        idx_vals = list(indexes.values())

        # either one or the other configuration (dependent vs. independent x/y axes)
        axis_dependent = (
            len(indexes) == 1
            and isinstance(idx_keys[0], tuple)
            and isinstance(idx_vals[0], CoordinateTransformIndex)
        )
        axis_independent = len(indexes) in (1, 2) and all(
            isinstance(idx, AxisAffineTransformIndex | PandasIndex) for idx in idx_vals
        )
        assert axis_dependent ^ axis_independent

        self._wrapped_indexes = dict(indexes)

    @classmethod
    def from_transform(
        cls, affine: Affine, width: int, height: int, x_dim: str = "x", y_dim: str = "y"
    ) -> RasterIndex:
        indexes: dict[WrappedIndexCoords, AxisAffineTransformIndex | CoordinateTransformIndex]

        # pixel centered coordinates
        affine = affine * Affine.translation(0.5, 0.5)

        if affine.is_rectilinear and affine.b == affine.d == 0:
            x_transform = AxisAffineTransform(affine, width, "x", x_dim, is_xaxis=True)
            y_transform = AxisAffineTransform(affine, height, "y", y_dim, is_xaxis=False)
            indexes = {
                "x": AxisAffineTransformIndex(x_transform),
                "y": AxisAffineTransformIndex(y_transform),
            }
        else:
            xy_transform = AffineTransform(affine, width, height, x_dim=x_dim, y_dim=y_dim)
            indexes = {("x", "y"): CoordinateTransformIndex(xy_transform)}

        return cls(indexes)

    @classmethod
    def from_variables(
        cls,
        variables: Mapping[Any, Variable],
        *,
        options: Mapping[str, Any],
    ) -> RasterIndex:
        # TODO: compute bounds, resolution and affine transform from explicit coordinates.
        raise NotImplementedError("Creating a RasterIndex from existing coordinates is not yet supported.")

    def create_variables(self, variables: Mapping[Any, Variable] | None = None) -> dict[Hashable, Variable]:
        new_variables: dict[Hashable, Variable] = {}

        for index in self._wrapped_indexes.values():
            new_variables.update(index.create_variables())

        return new_variables

    def isel(self, indexers: Mapping[Any, int | slice | np.ndarray | Variable]) -> RasterIndex | None:
        new_indexes: dict[WrappedIndexCoords, WrappedIndex] = {}

        for coord_names, index in self._wrapped_indexes.items():
            index_indexers = _filter_dim_indexers(index, indexers)
            if not index_indexers:
                # no selection to perform: simply propagate the index
                # TODO: uncomment when https://github.com/pydata/xarray/issues/10063 is fixed
                # new_indexes[coord_names] = index
                ...
            else:
                new_index = index.isel(index_indexers)
                if new_index is not None:
                    new_indexes[coord_names] = new_index

        if new_indexes:
            # TODO: if there's only a single PandasIndex can we just return it?
            # (maybe better to keep it wrapped if we plan to later make RasterIndex CRS-aware)
            return RasterIndex(new_indexes)
        else:
            return None

    def sel(self, labels: dict[Any, Any], method=None, tolerance=None) -> IndexSelResult:
        results = []

        for coord_names, index in self._wrapped_indexes.items():
            if not isinstance(coord_names, tuple):
                coord_names = (coord_names,)
            index_labels = {k: v for k, v in labels.items() if k in coord_names}
            if index_labels:
                results.append(index.sel(index_labels, method=method, tolerance=tolerance))

        return merge_sel_results(results)

    def equals(self, other: Index) -> bool:
        if not isinstance(other, RasterIndex):
            return False
        if set(self._wrapped_indexes) != set(other._wrapped_indexes):
            return False

        return all(
            index.equals(other._wrapped_indexes[k])  # type: ignore[arg-type]
            for k, index in self._wrapped_indexes.items()
        )

    def to_pandas_index(self) -> pd.Index:
        # conversion is possible only if this raster index encapsulates
        # exactly one AxisAffineTransformIndex or a PandasIndex associated
        # to either the x or y axis (1-dimensional) coordinate.
        if len(self._wrapped_indexes) == 1:
            index = next(iter(self._wrapped_indexes.values()))
            if isinstance(index, AxisAffineTransformIndex | PandasIndex):
                return index.to_pandas_index()

        raise ValueError("Cannot convert RasterIndex to pandas.Index")

    def __repr__(self):
        items: list[str] = []

        for coord_names, index in self._wrapped_indexes.items():
            items += [repr(coord_names) + ":", textwrap.indent(repr(index), "    ")]

        return "RasterIndex\n" + "\n".join(items)

    def transform(self) -> Affine:
        """Returns Affine transform for top-left corners."""
        if len(self._wrapped_indexes) > 1:
            x = self._wrapped_indexes["x"].axis_transform.affine
            y = self._wrapped_indexes["y"].axis_transform.affine
            aff = Affine(x.a, x.b, x.c, y.d, y.e, y.f)
        else:
            index = next(iter(self._wrapped_indexes.values()))
            aff = index.affine
        return aff * Affine.translation(-0.5, -0.5)



from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
import numpy as np
import xarray as xr
from odc.geo.geobox import GeoboxTiles

if TYPE_CHECKING:
    import dask.array
    import dask_geopandas


def tiles_to_array(tiles: GeoboxTiles) -> np.ndarray:
    shape = tiles.shape
    array = np.empty(shape=(shape.y, shape.x), dtype=object)
    for i in range(shape.x):
        for j in range(shape.y):
            array[j, i] = tiles[j, i]

    assert array.shape == tiles.shape
    return array


def is_in_memory(*, obj, geometries) -> bool:
    return not obj.chunks and isinstance(geometries, gpd.GeoDataFrame)


def geometries_as_dask_array(
    geometries: gpd.GeoDataFrame | dask_geopandas.GeoDataFrame,
) -> dask.array.Array:
    from dask.array import from_array

    if isinstance(geometries, gpd.GeoDataFrame):
        return from_array(geometries.geometry.to_numpy(), chunks=-1)
    else:
        divisions = geometries.divisions
        if any(d is None for d in divisions):
            print("computing current divisions, this may be expensive.")
            divisions = geometries.compute_current_divisions()
        chunks = np.diff(divisions).tolist()
        chunks[-1] += 1
        return geometries.to_dask_array(lengths=chunks).squeeze(axis=1)


def prepare_for_dask(
    obj: xr.Dataset | xr.DataArray,
    geometries,
    *,
    xdim: str,
    ydim: str,
    geoms_rechunk_size: int | None,
):
    from dask.array import from_array

    box = obj.odc.geobox

    chunks = (
        obj.chunksizes.get(ydim, obj.sizes[ydim]),
        obj.chunksizes.get(xdim, obj.sizes[ydim]),
    )
    tiles = GeoboxTiles(box, tile_shape=chunks)
    tiles_array = from_array(tiles_to_array(tiles), chunks=(1, 1))
    geom_array = geometries_as_dask_array(geometries)
    if geoms_rechunk_size is not None:
        geom_array = geom_array.rechunk({0: geoms_rechunk_size})
    return chunks, tiles_array, geom_array



from .raster_index import RasterIndex, assign_index


def _get_version():
    __version__ = "999"
    try:
        from ._version import __version__
    except ImportError:
        pass
    return __version__


__version__ = _get_version()

__all__ = ["RasterIndex", "assign_index"]



# rasterio wrappers
from __future__ import annotations

from collections.abc import Sequence
from functools import partial
from typing import TYPE_CHECKING, Any

import geopandas as gpd
import numpy as np
import odc.geo.xr  # noqa
import rasterio as rio
import xarray as xr
from rasterio.features import MergeAlg, geometry_mask
from rasterio.features import rasterize as rasterize_rio

from .utils import is_in_memory, prepare_for_dask

if TYPE_CHECKING:
    import dask_geopandas


def dask_rasterize_wrapper(
    geom_array: np.ndarray,
    tile_array: np.ndarray,
    offset_array: np.ndarray,
    *,
    fill: Any,
    all_touched: bool,
    merge_alg: MergeAlg,
    dtype_: np.dtype,
    env: rio.Env | None = None,
) -> np.ndarray:
    tile = tile_array.item()
    offset = offset_array.item()

    return rasterize_geometries(
        geom_array[:, 0, 0].tolist(),
        tile=tile,
        offset=offset,
        all_touched=all_touched,
        merge_alg=merge_alg,
        fill=fill,
        dtype=dtype_,
        env=env,
    )[np.newaxis, :, :]


def rasterize_geometries(
    geometries: Sequence[Any],
    *,
    dtype: np.dtype,
    tile,
    offset,
    env: rio.Env | None = None,
    clear_cache: bool = False,
    **kwargs,
):
    # From https://rasterio.readthedocs.io/en/latest/api/rasterio.features.html#rasterio.features.rasterize
    #    The out array will be copied and additional temporary raster memory equal to 2x the smaller of out data
    #    or GDAL’s max cache size (controlled by GDAL_CACHEMAX, default is 5% of the computer’s physical memory) is required.
    #    If GDAL max cache size is smaller than the output data, the array of shapes will be iterated multiple times.
    #    Performance is thus a linear function of buffer size. For maximum speed, ensure that GDAL_CACHEMAX
    #    is larger than the size of out or out_shape.
    if env is None:
        # out_size = dtype.itemsize * math.prod(tile.shape)
        # env = rio.Env(GDAL_CACHEMAX=1.2 * out_size)
        # FIXME: figure out a good default
        env = rio.Env()
    with env:
        res = rasterize_rio(
            zip(geometries, range(offset, offset + len(geometries)), strict=True),
            out_shape=tile.shape,
            transform=tile.affine,
            **kwargs,
        )
    if clear_cache:
        with rio.Env(GDAL_CACHEMAX=0):
            try:
                from osgeo import gdal

                # attempt to force-clear the GDAL cache
                assert gdal.GetCacheMax() == 0
            except ImportError:
                pass
    assert res.shape == tile.shape
    return res


def rasterize(
    obj: xr.Dataset | xr.DataArray,
    geometries: gpd.GeoDataFrame | dask_geopandas.GeoDataFrame,
    *,
    xdim="x",
    ydim="y",
    all_touched: bool = False,
    merge_alg: MergeAlg = MergeAlg.replace,
    geoms_rechunk_size: int | None = None,
    env: rio.Env | None = None,
) -> xr.DataArray:
    """
    Dask-aware wrapper around ``rasterio.features.rasterize``.

    Returns a 2D DataArray with integer codes for cells that are within the provided geometries.

    Parameters
    ----------
    obj: xr.Dataset or xr.DataArray
        Xarray object, whose grid to rasterize
    geometries: GeoDataFrame
        Either a geopandas or dask_geopandas GeoDataFrame
    xdim: str
        Name of the "x" dimension on ``obj``.
    ydim: str
        Name of the "y" dimension on ``obj``.
    all_touched: bool = False
        Passed to ``rasterio.features.rasterize``
    merge_alg: rasterio.MergeAlg
        Passed to ``rasterio.features.rasterize``.
    geoms_rechunk_size: int | None = None
        Size to rechunk the geometry array to *after* conversion from dataframe.
    env: rasterio.Env
        Rasterio Environment configuration. For example, use set ``GDAL_CACHEMAX`
        by passing ``env = rio.Env(GDAL_CACHEMAX=100 * 1e6)``.

    Returns
    -------
    DataArray
        2D DataArray with geometries "burned in"
    """
    if xdim not in obj.dims or ydim not in obj.dims:
        raise ValueError(f"Received {xdim=!r}, {ydim=!r} but obj.dims={tuple(obj.dims)}")
    box = obj.odc.geobox
    rasterize_kwargs = dict(all_touched=all_touched, merge_alg=merge_alg)
    # FIXME: box.crs == geometries.crs
    if is_in_memory(obj=obj, geometries=geometries):
        geom_array = geometries.to_numpy().squeeze(axis=1)
        rasterized = rasterize_geometries(
            geom_array.tolist(),
            tile=box,
            offset=0,
            dtype=np.min_scalar_type(len(geometries)),
            fill=len(geometries),
            env=env,
            **rasterize_kwargs,
        )
    else:
        from dask.array import from_array, map_blocks

        chunks, tiles_array, geom_array = prepare_for_dask(
            obj, geometries, xdim=xdim, ydim=ydim, geoms_rechunk_size=geoms_rechunk_size
        )
        # DaskGeoDataFrame.len() computes!
        num_geoms = geom_array.size
        # with dask, we use 0 as a fill value and replace it later
        dtype = np.min_scalar_type(num_geoms)
        # add 1 to the offset, to account for 0 as fill value
        npoffsets = np.cumsum(np.array([0, *geom_array.chunks[0][:-1]])) + 1
        offsets = from_array(npoffsets, chunks=1)

        rasterized = map_blocks(
            dask_rasterize_wrapper,
            geom_array[:, np.newaxis, np.newaxis],
            tiles_array[np.newaxis, :, :],
            offsets[:, np.newaxis, np.newaxis],
            chunks=((1,) * geom_array.numblocks[0], chunks[0], chunks[1]),
            meta=np.array([], dtype=dtype),
            fill=0,  # good identity value for both sum & replace.
            **rasterize_kwargs,
            dtype_=dtype,
        )
        if merge_alg is MergeAlg.replace:
            rasterized = rasterized.max(axis=0)
        elif merge_alg is MergeAlg.add:
            rasterized = rasterized.sum(axis=0)

        # and reduce every other value by 1
        rasterized = rasterized.map_blocks(partial(replace_values, to=num_geoms))

    return xr.DataArray(
        dims=(ydim, xdim),
        data=rasterized,
        coords=xr.Coordinates(
            coords={
                xdim: obj.coords[xdim],
                ydim: obj.coords[ydim],
                "spatial_ref": obj.spatial_ref,
                # TODO: figure out how to propagate geometry array
                # "geometry": geom_array,
            },
            indexes={xdim: obj.xindexes[xdim], ydim: obj.xindexes[ydim]},
        ),
        name="rasterized",
    )


def replace_values(array: np.ndarray, to, *, from_=0) -> np.ndarray:
    mask = array == from_
    array[~mask] -= 1
    array[mask] = to
    return array


# ===========> geometry_mask


def dask_mask_wrapper(
    geom_array: np.ndarray,
    tile_array: np.ndarray,
    *,
    all_touched: bool,
    invert: bool,
    env: rio.Env | None = None,
) -> np.ndarray[Any, np.dtype[np.bool_]]:
    tile = tile_array.item()

    return np_geometry_mask(
        geom_array[:, 0, 0].tolist(),
        tile=tile,
        all_touched=all_touched,
        invert=invert,
        env=env,
    )[np.newaxis, :, :]


def np_geometry_mask(
    geometries: Sequence[Any],
    *,
    tile,
    env: rio.Env | None = None,
    clear_cache: bool = False,
    **kwargs,
) -> np.ndarray[Any, np.dtype[np.bool_]]:
    # From https://rasterio.readthedocs.io/en/latest/api/rasterio.features.html#rasterio.features.rasterize
    #    The out array will be copied and additional temporary raster memory equal to 2x the smaller of out data
    #    or GDAL’s max cache size (controlled by GDAL_CACHEMAX, default is 5% of the computer’s physical memory) is required.
    #    If GDAL max cache size is smaller than the output data, the array of shapes will be iterated multiple times.
    #    Performance is thus a linear function of buffer size. For maximum speed, ensure that GDAL_CACHEMAX
    #    is larger than the size of out or out_shape.
    if env is None:
        # out_size = np.bool_.itemsize * math.prod(tile.shape)
        # env = rio.Env(GDAL_CACHEMAX=1.2 * out_size)
        # FIXME: figure out a good default
        env = rio.Env()
    with env:
        res = geometry_mask(geometries, out_shape=tile.shape, transform=tile.affine, **kwargs)
    if clear_cache:
        with rio.Env(GDAL_CACHEMAX=0):
            try:
                from osgeo import gdal

                # attempt to force-clear the GDAL cache
                assert gdal.GetCacheMax() == 0
            except ImportError:
                pass
    assert res.shape == tile.shape
    return res


def geometry_clip(
    obj: xr.Dataset | xr.DataArray,
    geometries: gpd.GeoDataFrame | dask_geopandas.GeoDataFrame,
    *,
    xdim="x",
    ydim="y",
    all_touched: bool = False,
    invert: bool = False,
    geoms_rechunk_size: int | None = None,
    env: rio.Env | None = None,
) -> xr.DataArray:
    """
    Dask-ified version of rioxarray.clip

    Parameters
    ----------
    obj : xr.DataArray | xr.Dataset
        Xarray object used to extract the grid
    geometries: GeoDataFrame | DaskGeoDataFrame
        Geometries used for clipping
    xdim: str
        Name of the "x" dimension on ``obj``.
    ydim: str
        Name of the "y" dimension on ``obj``
    all_touched: bool
        Passed to rasterio
    invert: bool
        Whether to preserve values inside the geometry.
    geoms_rechunk_size: int | None = None,
        Chunksize for geometry dimension of the output.
    env: rasterio.Env
        Rasterio Environment configuration. For example, use set ``GDAL_CACHEMAX`
        by passing ``env = rio.Env(GDAL_CACHEMAX=100 * 1e6)``.

    Returns
    -------
    DataArray
        3D dataarray with coverage fraction. The additional dimension is "geometry".
    """
    invert = not invert  # rioxarray clip convention -> rasterio geometry_mask convention
    if xdim not in obj.dims or ydim not in obj.dims:
        raise ValueError(f"Received {xdim=!r}, {ydim=!r} but obj.dims={tuple(obj.dims)}")
    box = obj.odc.geobox
    geometry_mask_kwargs = dict(all_touched=all_touched, invert=invert)

    if is_in_memory(obj=obj, geometries=geometries):
        geom_array = geometries.to_numpy().squeeze(axis=1)
        mask = np_geometry_mask(geom_array.tolist(), tile=box, env=env, **geometry_mask_kwargs)
    else:
        from dask.array import map_blocks

        chunks, tiles_array, geom_array = prepare_for_dask(
            obj, geometries, xdim=xdim, ydim=ydim, geoms_rechunk_size=geoms_rechunk_size
        )
        mask = map_blocks(
            dask_mask_wrapper,
            geom_array[:, np.newaxis, np.newaxis],
            tiles_array[np.newaxis, :, :],
            chunks=((1,) * geom_array.numblocks[0], chunks[0], chunks[1]),
            meta=np.array([], dtype=bool),
            **geometry_mask_kwargs,
        )
        mask = mask.any(axis=0)

    mask_da = xr.DataArray(
        dims=(ydim, xdim),
        data=mask,
        coords=xr.Coordinates(
            coords={
                xdim: obj.coords[xdim],
                ydim: obj.coords[ydim],
                "spatial_ref": obj.spatial_ref,
            },
            indexes={xdim: obj.xindexes[xdim], ydim: obj.xindexes[ydim]},
        ),
    )
    return obj.where(mask_da)
